//
//  ADUDriver.h
//
//  Created by Michael Dolinar on 2012-08-06.
//
//  Purpose:
//  Complete library for reading & writing to Ontrak Control Systems'
//  ADU I/O devices on Apple OSX 10.5 or newer.
//
//  Usage:
//  ADUInit(); gets things rolling
//
//  ADUCount(); make sure that your ADU device has been recognized
//
//  ADUWrite(pipe, command, ADUSerialNumber); send a command to a connected device
//                                            with a matching serial number
//
//  ADURead(pipe, ADUSerialNumber); read a response from a connected device with
//                                  a matching serial number
//
//  note: You may always send NULL instead of a specific ADU device serial number.
//        This will cause the first ADU device (at index 0) to be read / written.
//        Use this method only when your system has a single ADU connected.
//
//  ADUPrintList(); to print a list of all detected ADUs to stdout (great for debugging)
//  ADUGetSerialNumberFromIndex(index); to get the ADU device serial number at a
//                                           specific location in the list of ADUs.
//
//  note: Combine ADUCount() with ADUGetSerialNumberFromIndex(index) in a for loop
//        to retrieve the serial numbers of all connected ADU devices.
//
//  ADUGetNameFromSerialNumber(ADUSerialNumber); to get the device model name of the
//                                                    connected ADU device with a matching
//                                                    serial number.
//
//  Testing Summary:
//  Tested using an ADU 208 on a MacBook Pro running OSX 10.8 & Xcode 4.4.1.
//
//
//  Changelog:
//
//  Updated Aug 16, 2012 by M. Dolinar
//  Version 0.4
//
//  - Replaced matching dictionary manufacturer name with vendor ID
//    to support Ontrak manufactured devices with non-standard
//    manufacturer names.
//  - Increased buffer sizes from 8 to 64 bytes.
//  - Modified ADUWrite to accept inputs up to 63 bytes (64 - 1).
//
//  Updated Aug 13, 2012 by M. Dolinar
//  Version 0.3
//
//  - Added ADUClose() to close all ADU devices (call when exiting your app)
//
//  Updated Aug 11, 2012 by M. Dolinar
//  Version 0.2
//
//  - Corrected serial number search bug
//  - Added serial number & device name return functions
//

#include <stdio.h>
#include <stdlib.h>
#include "ADUDriver.h"
#include <IOKit/hid/IOHIDManager.h>





// launches the IOHIDManager for ADU devices
//
bool ADUInit(void){
    rootADUNode = NULL;
    
    // Create the USB HID Manager (OSX 10.5 & later only)
    IOHIDManagerRef HIDManager = IOHIDManagerCreate(kCFAllocatorDefault,
                                                    kIOHIDOptionsTypeNone);
    
    // Create an empty matching dictionary for filtering USB devices in our HID manager
    CFMutableDictionaryRef matchDict = CFDictionaryCreateMutable(kCFAllocatorDefault,
                                                                 2,
                                                                 &kCFTypeDictionaryKeyCallBacks,
                                                                 &kCFTypeDictionaryValueCallBacks);
    
    // Specify the USB device manufacturer (by vendor ID) in our matching dictionary
    int vendorId = 0x0A07;
    CFNumberRef vendorIDCFNumRef  = CFNumberCreate( kCFAllocatorDefault, kCFNumberIntType, &vendorId );

    CFDictionarySetValue(matchDict,
                         CFSTR(kIOHIDVendorIDKey),
                         vendorIDCFNumRef);
    
    // Apply the matching to our HID manager
    IOHIDManagerSetDeviceMatching(HIDManager, matchDict);
    
    // We're done with the matching dictionary
    if(matchDict) CFRelease(matchDict);
    
    // The HID manager will use callbacks when specified USB devices are connected / disconnected.
    // We just need to register the callback functions (here)
    IOHIDManagerRegisterDeviceMatchingCallback(HIDManager, &Handle_DeviceMatchingCallback, NULL);
    IOHIDManagerRegisterDeviceRemovalCallback(HIDManager, &Handle_DeviceRemovalCallback, NULL);
    
    // Add the HID manager to the main run loop (or the callback functions won't be called)
    IOHIDManagerScheduleWithRunLoop(HIDManager, CFRunLoopGetMain(), kCFRunLoopDefaultMode);
    
    // Try to open the HID mangager
    IOReturn IOReturn = IOHIDManagerOpen(HIDManager, kIOHIDOptionsTypeNone);
    if(IOReturn) {
        printf("IOHIDManagerOpen failed.\n");  //  Couldn't open the HID manager!
    
        // TODO: application specific error handling
        
        return FALSE;
    }
    
    return TRUE;
}





// Writes command to the ADU device's command channel of the device with serial number ADUSNo.
// Returns sucess.
// note: if ADUSNo is null, writes to first opened ADU device (usefull for
// applications where only one ADU will every be connected.)
//
bool ADUWrite(int pipe, char* command, char* ADUSerialNumber){
    
    // initialize the output buffer. 8 bytes - first byte set to specifies the pipe
    char* outBuffer = calloc(64,1);
    outBuffer[0] = pipe;
    
    // get the command length (max length is 7 bytes)
    int commandLength = 63;
    if(strlen(command) < 63) commandLength = (int) strlen(command);
    
    // put the command into the output buffer
    for (int x = 0; x < commandLength; x++) outBuffer[x+1] = command[x];
    
    // initialize a pointer to the target device (if available)
    struct ADUDeviceNode* targetDevice = ADUFindNodeFromSerialNumber(ADUSerialNumber);
    
    if(targetDevice) {
        // send a device report
        IOReturn  tIOReturn = IOHIDDeviceSetReport(*targetDevice->reference, // IOHIDDeviceRef for the HID device
                                                   kIOHIDReportTypeOutput,   // IOHIDReportType for the report
                                                   0,                        // CFIndex for the report ID
                                                   (uint8_t*)outBuffer,      // address of report buffer
                                                   sizeof(outBuffer));       // length of the report
     
        if(tIOReturn) {
            printf("SetReport return value: %08x\n", tIOReturn);             // log error
            
            // TODO: application specific error handling
            
            return FALSE;                                                    // device write failed
        } else {
            return TRUE;                                                     // device write suceeded
        }
    }
    
    return FALSE; // no ADU devices OR specified serial number not found
}





// Return a pointer to the specified device's read buffer.
// Pipe 1 (command), 2 (RS-232) or 3 (streaming) must be specified.
// ADUSNo may be NULL if the first device's buffer is needed (usefull
// for applications where only one ADU will every be connected.)
//
char* ADURead(int pipe, char* ADUSerialNumber){
    struct ADUDeviceNode* targetDevice = ADUFindNodeFromSerialNumber(ADUSerialNumber);
    
    if(targetDevice) {  // if device found, return specified input buffer
        switch (pipe) {
            case 1: return targetDevice->pipe1Buffer;
            case 2: return targetDevice->pipe2Buffer;
            case 3: return targetDevice->pipe3Buffer;
        }
        
    }
    
    return NULL;    // no ADU devices OR specified serial number not found
}





// Counts the ADUDeviceNodes in the linked list of ADU devices
//
int ADUCount(void){
    
    int count = 0;                                      // initialize the counter
    struct ADUDeviceNode* currentNode = rootADUNode;    // for traversing the list
        
    while(currentNode){                                 // until the last node
        count++;                                        // count this node
        currentNode = currentNode->next;                // move to the next node
    }                                                   
    
    return count;
}





// Prints the current list to stdout. Used for debugging, left in for your pleasure.
//
void ADUPrintList(void){
    
    printf("ADU List: {");                              // print headder
    
    struct ADUDeviceNode* currentNode = rootADUNode;    // for traversing the list
    int count = 0;                                      // init the counter
        
    while(currentNode){                                 // traverse the list, print each node, enumerate
        if(count > 0) printf(", ");                     // seperate elements by a comma
        count++;                                        // increment the counter
            
        printf("%i: %s / %s",                           // print element
                count,
                currentNode->name,
                currentNode->serialNumber);
            
        currentNode = currentNode->next;                // next node
    }
    
    printf("}\n");                                      // print footer
}





// returns a pointer to the ADU at index
//
char* ADUGetSerialNumberFromIndex(int index){
    
    struct ADUDeviceNode* currentNode = rootADUNode;    // for traversing the list
    int count = 0;                                      // init the counter
    
    while(currentNode){                                 // traverse the list, print each node, enumerate
        
        if(count == index)                              // if this is the specified index...
            return currentNode->serialNumber;           // ... return a pointer to the s/no
        
        count++;
        currentNode = currentNode->next;                // next node
    }
    return NULL;                                        // index not found / nothing to return
}





// returns a pointer to the ADU name with s/no ADUSNo
//
char* ADUGetNameFromSerialNumber(char* ADUSerialNumber){
    
    // initialize a pointer to the target device (if available)
    struct ADUDeviceNode* targetDevice = ADUFindNodeFromSerialNumber(ADUSerialNumber);
    
    if(targetDevice) return targetDevice->name;     // if we have a device, return the name
    return NULL;                                    // otherwise, device wasn't found
}





// Closes all ADUs. Call before exiting your program.
//
void ADUClose(void){
    struct ADUDeviceNode* currentNode = rootADUNode;    // for traversing the list
    
    while(currentNode){                                 // until the last node
        IOHIDDeviceClose(*currentNode->reference,
                         kIOHIDOptionsTypeNone);        // close the device
        currentNode = currentNode->next;                // move to the next node
    }
}





// Adds an ADUDeviceNode to the linked list. Called by the 'device added' callback
//
static struct ADUDeviceNode* ADUAddNode(IOHIDDeviceRef* newDevice){
    struct ADUDeviceNode* currentNode = rootADUNode;            // for traversing the list
    
    // retrieve the device name & serial number (these are strings when using with ADU devices)
    char* newName = (char*) CFStringGetCStringPtr(IOHIDDeviceGetProperty(*newDevice, CFSTR("Product")), kCFStringEncodingMacRoman);
    char* newSNo = (char*) CFStringGetCStringPtr(IOHIDDeviceGetProperty(*newDevice, CFSTR("SerialNumber")), kCFStringEncodingMacRoman);
    
    while(currentNode) currentNode = currentNode->next;         // find the end of the list
    
    // Create a new node at the end of the list
    if(rootADUNode) currentNode = (struct ADUDeviceNode*) calloc(sizeof(struct ADUDeviceNode), 1);
    else {
        rootADUNode = (struct ADUDeviceNode*) calloc(sizeof(struct ADUDeviceNode), 1);
        currentNode = rootADUNode;
    }
    
    // populate the new node
    currentNode->name = calloc(strlen(newName),1);              // init the name field
    strcpy(currentNode->name, newName);                         // copy the device name
    
    currentNode->serialNumber = calloc(strlen(newSNo),1);       // init the serial number
    strcpy(currentNode->serialNumber, newSNo);                  // copy the device s/no
    
    currentNode->pipe1Buffer = calloc(64,1);                     // init the input buffers
    currentNode->pipe2Buffer = calloc(64,1);
    currentNode->pipe3Buffer = calloc(64,1);
    currentNode->readBuffer = calloc(64,1);
    
    currentNode->reference = calloc(sizeof(IOHIDDeviceRef), 1); // init the device reference
    *currentNode->reference = *newDevice;                       // assign the device ref

    return currentNode;
}





// Removes specified node from the list. Called by the 'device removed' callback
//
static bool ADUKillNode(IOHIDDeviceRef* killDevice){
    struct ADUDeviceNode* currentNode = rootADUNode;    // for traversing the list
    struct ADUDeviceNode* previousNode = NULL;          // for bridging the gap when a link is removed
    
    while(currentNode) {                                // traverse the list
        if (*currentNode->reference == *killDevice) {   // is this our target device?
                
            if(previousNode) previousNode->next = currentNode->next;
            else rootADUNode = currentNode->next;
            
            free(currentNode->name);                    // deallocate this node
            free(currentNode->serialNumber);
            free(currentNode->pipe1Buffer);
            free(currentNode->pipe2Buffer);
            free(currentNode->pipe3Buffer);
            free(currentNode->readBuffer);
            free(currentNode->reference);
            free(currentNode);
            
            return true;                                // done.
        }
            
        previousNode = currentNode;                     // progress to the next node
        currentNode = currentNode->next;
    }

    return false;                                       // target device not found.
}





// Searches linked list of ADU devices for specified serial number. Returns pointer
// to the structure with matching serial number, or NULL.
// note: if serialNumber is NULL, returns pointer to root (first ADU opened.)
//
static struct ADUDeviceNode* ADUFindNodeFromSerialNumber(char* serialNumber){
    
    if(serialNumber) {      // Serial number specified, try to find it.
        struct ADUDeviceNode* currentNode = rootADUNode;    // for traversing the list
        
        while(currentNode){                                 // until the last node
            if(*currentNode->serialNumber == *serialNumber) return currentNode;
            currentNode = currentNode->next;                // move to the next node
        }
    } else {                // No serial number was specified,
        return rootADUNode; // just return root.
    }
    
    return NULL; // Specified serial number not found. 
}





// Device specified in the matching dictionary has been added (callback function)
//
static void Handle_DeviceMatchingCallback(void *inContext,
                                          IOReturn inResult,
                                          void *inSender,
                                          IOHIDDeviceRef inIOHIDDeviceRef){

    IOReturn o = IOHIDDeviceOpen(inIOHIDDeviceRef, kIOHIDOptionsTypeSeizeDevice);
    
    if(o == kIOReturnSuccess) {                                         // if device opened
        
        struct ADUDeviceNode* newADU = ADUAddNode(&inIOHIDDeviceRef);   // add it to the list
        
        IOHIDDeviceRegisterInputReportCallback(*newADU->reference,      // register input callback
                                               (uint8_t*) newADU->readBuffer,
                                               sizeof(newADU->readBuffer),
                                               Handle_IOHIDDeviceInputReportCallback,
                                               newADU);
    
    }
    else {
        printf("Open Failed: %i\n",o);                                  // log the error
        
        // TODO: application-specific error handling
    }
}





// Device specified in the matching dictionary has been removed (callback function)
//
static void Handle_DeviceRemovalCallback(void *inContext,
                                         IOReturn inResult,
                                         void *inSender,
                                         IOHIDDeviceRef inIOHIDDeviceRef){
    
    IOHIDDeviceRegisterInputReportCallback(inIOHIDDeviceRef,    // de-register input callback
                                           NULL,
                                           0,
                                           NULL,
                                           NULL);
    
    ADUKillNode(&inIOHIDDeviceRef);                             // remove the device from the linked list
}





// Called whenever data is recieved from the ADU device
//
static void Handle_IOHIDDeviceInputReportCallback(void * inContext,
                                                  IOReturn inResult,
                                                  void * inSender,
                                                  IOHIDReportType inType,
                                                  uint32_t inReportID,
                                                  uint8_t * inReport,
                                                  CFIndex inReportLength){
    if (inResult) {
        printf("%08x\n", inResult);
        
        // TODO: applicaiton-specific error handling
        
        return;
    }
    // retrieve the device
    struct ADUDeviceNode* respondingDevice = inContext;
    
    // copy the input report to the applicable buffer
    if ((strlen((char*) inReport) > 1) && (respondingDevice)) {
        
        // extract the data packet
        char* dataPacket = calloc(7,1);
        for (int x = 0; x < 7; x++) if (x < inReportLength) dataPacket[x] = inReport[x + 1];
        
        // store the report in the correct buffer
        switch (inReport[0]) {
            case 1: strcpy(respondingDevice->pipe1Buffer, dataPacket);
            case 2: strcpy(respondingDevice->pipe2Buffer, dataPacket);
            case 3: strcpy(respondingDevice->pipe3Buffer, dataPacket);
        }
    }
}




