//
//  ADUDriver.h
//
//  Created by Michael Dolinar on 2012-08-06.
//
//  Purpose:
//  Complete library for reading & writing to Ontrak Control Systems'
//  ADU I/O devices on Apple OSX 10.5 or newer.
//
//  Usage:
//  ADUInit(); gets things rolling
//
//  ADUCount(); make sure that your ADU device has been recognized
//
//  ADUWrite(pipe, command, ADUSerialNumber); send a command to a connected device
//                                            with a matching serial number
//
//  ADURead(pipe, ADUSerialNumber); read a response from a connected device with
//                                  a matching serial number
//
//  note: You may always send NULL instead of a specific ADU device serial number.
//        This will cause the first ADU device (at index 0) to be read / written.
//        Use this method only when your system has a single ADU connected.
//
//  ADUPrintList(); to print a list of all detected ADUs to stdout (great for debugging)
//  ADUGetSerialNumberFromIndex(index); to get the ADU device serial number at a 
//                                           specific location in the list of ADUs.
//
//  note: Combine ADUCount() with ADUGetSerialNumberFromIndex(index) in a for loop
//        to retrieve the serial numbers of all connected ADU devices.
//
//  ADUGetNameFromSerialNumber(ADUSerialNumber); to get the device model name of the 
//                                                    connected ADU device with a matching
//                                                    serial number.
//
//  Testing Summary:
//  Tested using an ADU 208 on a MacBook Pro running OSX 10.8 & Xcode 4.4.1.
//
//
//  Changelog:
//
//  Updated Aug 16, 2012 by M. Dolinar
//  Version 0.4
//
//  - Replaced matching dictionary manufacturer name with vendor ID
//    to support Ontrak manufactured devices with non-standard
//    manufacturer names.
//  - Increased buffer sizes from 8 to 64 bytes.
//  - Modified ADUWrite to accept inputs up to 63 bytes (64 - 1).
//
//  Updated Aug 13, 2012 by M. Dolinar
//  Version 0.3
//
//  - Added ADUClose() to close all ADU devices (call when exiting your app)
//
//  Updated Aug 11, 2012 by M. Dolinar
//  Version 0.2
//
//  - Corrected serial number search bug
//  - Added serial number & device name return functions
//

#ifndef ADUDriver_h
#define ADUDriver_h

#include <IOKit/hid/IOHIDManager.h>





// All of the information required to maintain a linked list of ADU devices
struct ADUDeviceNode {
    char*                   name;           // device's name as a string
    char*                   serialNumber;   // device's serial number as a string
    char*                   readBuffer;     // read buffer (8 bytes) written by the HIDManager
    char*                   pipe1Buffer;    // control pipe buffer (8 bytes)
    char*                   pipe2Buffer;    // serial pipe buffer (8 bytes)
    char*                   pipe3Buffer;    // streaming buffer (8 bytes)
    IOHIDDeviceRef*         reference;      // device ref for use with the IOHIDManager
    struct ADUDeviceNode*   next;           // next node, or NULL if last node
};

struct ADUDeviceNode* rootADUNode;   // root node for the linked list of ADU devices





// ********************************************************* Externally available interface *****
// *
bool  ADUInit(void);                            // launches the IOHIDManager for ADU devices
bool  ADUWrite(int pipe, char* command, char* ADUSerialNumber);
                                                // writes a command to the ADU with s/no ADUSNo
char* ADURead(int pipe, char* ADUSerialNumber); // returs a pointer to the specified input buffer
int   ADUCount(void);                           // counts the available nodes
void  ADUPrintList(void);                       // prints the current list to stdout
char* ADUGetSerialNumberFromIndex(int index);   // returns a pointer to the ADU at index
char* ADUGetNameFromSerialNumber(char* ADUSerialNumber);
                                                // returns a pointer to the ADU name with s/no ADUSNo
void ADUClose(void);                            // Closes all open ADUs





// ********************************************************** Internal functions ***************
// *
static struct ADUDeviceNode* ADUAddNode(IOHIDDeviceRef* newDevice);             // adds a node to the list
static bool ADUKillNode(IOHIDDeviceRef* killDevice);                            // removes specified node from the list
static struct ADUDeviceNode* ADUFindNodeFromSerialNumber(char* serialNumber);   // returns NULL or pointer to target ADU

// USB device added callback function
static void Handle_DeviceMatchingCallback(void *inContext,
                                          IOReturn inResult,
                                          void *inSender,
                                          IOHIDDeviceRef inIOHIDDeviceRef);

// USB device removed callback function
static void Handle_DeviceRemovalCallback(void *inContext,
                                         IOReturn inResult,
                                         void *inSender,
                                         IOHIDDeviceRef inIOHIDDeviceRef);

// Report arrived
static void Handle_IOHIDDeviceInputReportCallback(void * inContext,
                                                  IOReturn inResult,
                                                  void * inSender,
                                                  IOHIDReportType inType,
                                                  uint32_t inReportID,
                                                  uint8_t * inReport,
                                                  CFIndex inReportLength);

#endif
